"""Package containing tools to notify in various ways.
"""

import typing
import logging

import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText


class BasicNotifier:
    """Basic notifier.

Sub-classes should inherit and implement `send` method
    """

    def __init__(self, config):
        self.config = config

    def send(self, subject: str, msg: str):
        """Send a message using the notifier.

        :param subject:  String subject (one line summary)

        :param msg:  String message to send.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Basic send message. Sub-classes must override.

        """
        raise NotImplementedError


class EchoNotifier(BasicNotifier):
    """Sub-class of notifier to just print to stdout.
    """

    def send(self, subject: str, msg: str):
        logging.debug('Using notifier %s', str(self))
        print('%s\n%s' % (subject, msg))


class LogInfoNotifier(BasicNotifier):
    """Sub-class of notifier to use logging.info to notify.
    """

    def send(self, subject: str, msg: str):
        logging.debug('Using notifier %s', str(self))
        logging.info('%s\n%s', subject, msg)


class SentryNotifier(BasicNotifier):
    """Sub-class of notifier to use to notify via sentry.
    """

    def send(self, subject: str, msg: str):
        logging.debug('Will capture sentry message for subject: %s', subject)
        dsn = self.config.sentry
        if not dsn:
            raise ValueError('No value for sentry dsn; %s' % (
                'did you forget --SENTRY <dsn> ?'))

        # pytype gets confused by conditional imports so don't do them
        # if we are in type checking mode
        if not typing.TYPE_CHECKING:
            # pylint: disable=import-outside-toplevel
            import sentry_sdk  # pylint: disable=import-error
            sentry_sdk.init(dsn)
            sentry_sdk.capture_message('%s\n%s' % (subject, msg))
        else:
            raise ValueError('Should not hit TYPE_CHECKING at runtime')
        logging.debug('Captured sentry message for subject: %s', subject)


class EmailNotifier(BasicNotifier):
    """Sub-class of notifier to send messages via email.
    """

    def send(self, subject, msg):
        hits = 0
        email_to = self.config.ox_mon_email_to
        if email_to:
            email_to = email_to.split(',')
        email_from = self.config.ox_mon_email_from
        profile = self.config.ox_mon_ses_profile
        if profile is not None:
            self.send_via_ses(subject, msg, profile, email_to, email_from)
            hits += 1
        gmail_passwd = self.config.ox_mon_gmail_passwd
        if gmail_passwd:
            self.send_via_gmail(subject, msg, gmail_passwd, email_to,
                                email_from)
            hits += 1
        if not hits:
            raise ValueError('Could not notify via email; no credentials')

    def send_via_gmail(self, subject: str, msg: str,
                       gmail_passwd: str, email_to: typing.List[str],
                       email_from: str):
        """Send email using gmail as a realy.

        :param subject:    String subject to send.

        :param msg:        String message to send.

        :param gmail_passwd:   String gmail password for email_from account.

        :param email_to:       List of strings to send email to.

        :param email_from:     String gmail account we send from.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Send email using gmail as SMTP relay.

        """
        logging.debug('Notifying from send_via_gmail: %s', str(self))
        gmail_msg = MIMEMultipart()
        gmail_msg['From'] = email_from
        gmail_msg['To'] = ', '.join(email_to)
        gmail_msg['Subject'] = subject
        gmail_msg.attach(MIMEText(msg))

        server = smtplib.SMTP('smtp.gmail.com', 587)
        server.starttls()
        server.login(email_from, gmail_passwd)
        server.sendmail(gmail_msg['From'], gmail_msg['To'],
                        gmail_msg.as_string())

    def send_via_ses(self, subject: str, msg: str, profile_name: str,
                     email_to: typing.List[str], email_from: str, **kwargs):
        """Send email using AWS SES.

        :param subject:    String subject to send.

        :param msg:        String message to send.

        :param profile_name:  AWS profile name to use.

        :param email_to:       List of strings to send email to.

        :param email_from:     String gmail account we send from.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Send email using AWS SES.

        """
        logging.debug('Notifying from send_via_ses: %s', str(self))
        try:
            # pylint: disable=import-outside-toplevel
            import boto3
        except Exception as problem:  # pylint: disable=broad-except
            logging.error('Unable to import boto3: %s\nIs boto3 installed?',
                          str(problem))
            raise
        dev = boto3.Session(profile_name=profile_name, **kwargs)
        client = dev.client('ses')
        result = client.send_email(Destination={
            'ToAddresses': email_to}, Message={
                'Body': {'Text':
                         {'Data': msg}},
                'Subject': {'Data': subject}}, Source=email_from)
        assert result is not None


# Dictionary of notifier names and corresponding classes.
# Should only be used by make_notifier.
_NDICT = {
    'echo': EchoNotifier,
    'email': EmailNotifier,
    'loginfo': LogInfoNotifier,
    'sentry': SentryNotifier
}


def make_notifier(ntype: str, config) -> BasicNotifier:
    """Make instance of BasicNotifier for given ntype.

    :param ntype:    String in _NDICT indicating notifier type (e.g.,
                     'echo' or 'email').

    :param config:   Instance of checker config.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  Instance of BasicNotifier for given ntype.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  Create notifier based on string type.

    """
    ncls = _NDICT[ntype]
    obj = ncls(config)
    return obj



"""Command line interface to ox_mon.
"""

import traceback
import sys
import logging
import typing

import click

from ox_mon import VERSION
from ox_mon.common import configs, exceptions
from ox_mon.checking import (
    apt_checker, clamav_checker, disk_checker, file_checker,
    version_checker)
from ox_mon.triggers import file_triggers

from ox_mon.misc import cmds


def prep_sentry(dsn):
    """Prepare sentry if a dsn is passed in.

We do an import inside this function so we do not try to import sentry
if no DSN is given (e.g., if sentry is not installed).
    """
    capture = {'name': 'unknown', 'func': logging.critical}

    # pytype gets confused by conditional imports so don't do them
    # if we are in type checking mode
    if not typing.TYPE_CHECKING:
        # pylint: disable=import-outside-toplevel
        import sentry_sdk  # pylint: disable=import-error
        sentry_sdk.init(dsn)
        capture = {'name': 'sentry', 'func': sentry_sdk.capture_exception}
    return capture


@click.group()
def main():
    "Command line interface to ox_mon."


@main.group()
def check():
    "Checking commands."


@main.group()
def trigger():
    "Trigger commands."


@main.group()
def gcmd():
    "General commands."


def add_options(olist: typing.List[configs.OxMonOption]):
    """Add options to a click command

    :param olist:  List of OxMonOption values.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    :return:  A decorator function to apply to a click command.

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:  For all elements in olist decorator to put those
              into a click command.
    """
    def _add_options(func):
        "Make decorator to apply options to click command."
        for opt in reversed(olist):
            func = click.option('--' + opt.key.lstrip('-'), **opt.data)(func)
        return func
    return _add_options


def generic_command(task_cls: typing.Callable, sentry: str,
                    loglevel: str, **kwargs):
    """

    :param task_cls:       Subclass of interface.OxMonTask to make task class

    :param sentry:         String DSN for sentry or None if not using sentry.

    :param loglevel:       String loglevel (e.g., 'DEBUG', 'INFO', etc.)

    :param **kwargs:       Passed to configs.BasicConfig

    ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

    PURPOSE:    Implement boilerplate for command line call of basic
                task. Basically, we do the following.

         1. Setup logging based on loglevel.
         2. Prepare sentry if desired.
         3. Create BasicConfig instance from **kwargs.
         4. Create task_cls instance using config.
         5. Run run method and if an exception is encountered, then
            capture with sentry if desired and return non-zero exit code.



    """
    root_logger = logging.getLogger('')
    root_logger.setLevel(getattr(logging, loglevel))
    capture = {'name': 'logging.critical', 'func': logging.critical}
    if sentry:
        capture = prep_sentry(sentry)
    try:
        config = configs.BasicConfig(sentry=sentry, **kwargs)
        task = task_cls(config)
        task.run()
    except exceptions.OxMonAlarm as ox_alarm:
        capture['func'](ox_alarm)
        sys.exit(1)
    except Exception as problem:  # pylint: disable=broad-except
        msg = 'Got Problem: %s' % str(problem)
        if capture:
            msg += '\nWill try to report via %s' % (
                capture.get('name', 'unknown'))
        logging.error(msg)
        capture['func'](problem)
        print(traceback.format_exc())
        sys.exit(2)


@check.command()
@add_options(apt_checker.AptChecker.options())
def apt(sentry, **kwargs):
    "Check to see what needs to be updated via 'apt'."

    return generic_command(apt_checker.AptShellChecker, sentry, **kwargs)


@check.command()
@add_options(disk_checker.SimpleDiskChecker.options())
def disk(sentry, **kwargs):
    "Check to see what needs to be updated via 'apt'."

    return generic_command(disk_checker.SimpleDiskChecker, sentry, **kwargs)


@check.command()
@add_options(clamav_checker.ClamScanShellChecker.options())
def clamscan(sentry, **kwargs):
    "Check virus scan using clamscan."

    return generic_command(clamav_checker.ClamScanShellChecker,
                           sentry, **kwargs)


@check.command()
@add_options(file_checker.SimpleFileChecker.options())
def filestatus(sentry, **kwargs):
    "Check file status."

    return generic_command(file_checker.SimpleFileChecker,
                           sentry, **kwargs)


@check.command()
@add_options(version_checker.SimpleVersionChecker.options())
def vcmp(sentry, **kwargs):
    "Compare version of some other command"

    return generic_command(version_checker.SimpleVersionChecker,
                           sentry, **kwargs)


@trigger.command()
@add_options(file_triggers.FileWatchCopy.options())
def fwatch(sentry, **kwargs):
    "Watch files in directory and copy them to archive."

    return generic_command(file_triggers.FileWatchCopy,
                           sentry, **kwargs)


@main.command()
def version():
    "Show version of ox_mon."
    msg = 'ox_mon version: %s' % VERSION
    click.echo(msg)
    return msg


@gcmd.command()
@add_options(cmds.RawCmd.options())
def raw(sentry, **kwargs):
    """Run raw command using ox_mon notification.

This command is helpful in case you want to run a shell script
or other basic command and use ox_mon to verify that it ran correclty
with notifications sent if there are problems.
    """

    return generic_command(cmds.RawCmd, sentry, **kwargs)


if __name__ == '__main__':
    main()



"""Module for trigger on file related events.
"""

import tempfile
import shutil
import hashlib
import datetime
import os
import logging
import types

from ox_mon.common import configs, interface, stubs

try:
    import inotify.adapters
    import inotify.constants
except Exception as problem:  # pylint: disable=broad-except
    inotify = types.SimpleNamespace(  # pylint: disable=invalid-name
        adapters=stubs.GenericModuleStub(
            'adapters', str(problem), 'inotify'))
    inotify.adapters.Inotify = inotify.adapters.make_complainer('Inotify')


class FileWatchCopy(interface.OxMonTask):
    """Trigger to copy files when they are created or modified.
    """

    @classmethod
    def options(cls):
        logging.debug('Making options for %s', cls)
        result = configs.BASIC_OPTIONS + [
            configs.OxMonOption(
                'watch', default=None, required=True, help=(
                    'Directory to watch for changes.')),
            configs.OxMonOption(
                'archive', default=None, required=True, help=(
                    'Directory to copy changes to.')),
            ]
        return result

    def _loop(self):
        """Check disk usage.
        """

        for key, name in [('archive', self.config.archive),
                          ('watch', self.config.watch)]:
            if not (name and os.path.exists(name)):
                raise ValueError('Directory %s = "%s" must exists.' % (
                    key, name))
            if not os.path.isdir(name):
                raise ValueError('%s %s must be a directory' % (
                    name, key))

        inote = inotify.adapters.InotifyTree(self.config.watch, (
            inotify.constants.IN_CLOSE_WRITE |
            inotify.constants.IN_CREATE |
            inotify.constants.IN_MOVED_TO |
            inotify.constants.IN_MOVE_SELF))

        for event in inote.event_gen(yield_nones=False):
            (header, type_names, path, filename) = event

            logging.debug('HEADER=%s', str(header))
            logging.info("PATH=[%s] FILENAME=[%s] EVENT_TYPES=%s",
                         path, filename, type_names)
            my_fname = os.path.join(path, filename)
            if os.path.isdir(my_fname):
                logging.info('Skip directory %s', my_fname)
                continue
            my_fd, tname = tempfile.mkstemp()
            os.close(my_fd)
            logging.info('Copying %s to temp file %s', my_fname, tname)
            shutil.copy(my_fname, tname)  # copy to tmp file first for safety
            hash_dir, my_utc = self.prep_destination(tname)
            if os.path.exists(os.path.join(hash_dir, 'main.data')):
                logging.info('main data already exists in dir %s', hash_dir)
            else:
                logging.info('Copying %s to %s', tname,
                             os.path.join(hash_dir, 'main.data'))
                shutil.copy(tname, os.path.join(hash_dir, 'main.data'))

            with open(os.path.join(hash_dir, 'names.txt'), 'a') as my_fd:
                my_fd.write('%s: %s\n' % (my_utc, my_fname))
                logging.info('At utc %s saved %s', my_utc, my_fname)
            os.remove(tname)

    def prep_destination(self, tname: str):
        """Make destination to store file and return hash and utc timestamp.

        :param tname:    String name for path to file.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  The pair (utc, hash_dir) where utc is a string timestamp
                  based on UTC time and hash_dir is a path to the directory
                  to store files with the hash for tname (hash_dir is
                  created if necessary).

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Prepare a place to store the file.

        """
        my_utc, my_hash = self.make_utc_and_hash(tname)
        hash_dir = os.path.join(self.config.archive, my_hash)
        if not os.path.exists(hash_dir):
            os.mkdir(hash_dir)
        return hash_dir, my_utc

    @staticmethod
    def make_utc_and_hash(fname: str):
        """Make and return utc timestamp and hash string.

        :param fname:    String name for path to file.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:  The pair (utc, hash) where utc is a string timestamp
                  based on UTC time and hash is the md5 hash of the file.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Get UTC timestamp and hash of file.

        """
        utc = datetime.datetime.utcnow().strftime('%Y%m%d_%H%M%S')
        md5 = hashlib.md5()
        with open(fname, 'rb') as my_fd:
            while True:
                data = my_fd.read(65536)
                if not data:
                    break
                md5.update(data)
        return utc, md5.hexdigest()

    def _do_task(self):
        return self._loop()



"""Common configuration information.
"""

import os

import click


class OxMonOption:
    """Option for running ox_mon
    """

    def __init__(self, key, **kwargs):
        self.key = key
        self.data = dict(kwargs)


BASIC_OPTIONS = [
    OxMonOption('--notifier', default=['echo'], multiple=True, help=(
        'Notifier(s) to use (e.g., echo or email or sentry). You can '
        'provide multiple versions of this if you want multiple '
        'notifiers (e.g., --notifier echo --notifier email). '
        'Some notifiers will need additional parameters such as '
        'the DSN (provided with --SENTRY) or email credentials.')),
    OxMonOption('--SENTRY', default=lambda: os.getenv('SENTRY_DSN'),
                help=('Optional Sentry DSN if you want error reporting '
                      'via sentry.')),
    OxMonOption(
        '--OX_MON_EMAIL_TO', default=None, envvar='OX_MON_EMAIL_TO', help=(
            'Dynamic option for email address to notify. '
            'Should be a comma separated group of email '
            'addresses (e.g., foo@examle.com or a@b.c,d@e.f).'
            'If not set, will lookup from environment.')),
    OxMonOption(
        '--OX_MON_EMAIL_FROM', default=None, envvar='OX_MON_EMAIL_FROM', help=(
            'Dynamic option for email address to notify from. '
            'If not set, will lookup from environment.')),
    OxMonOption('--OX_MON_GMAIL_PASSWD', default=None,
                envvar='OX_MON_GMAIL_PASSWD', help=(
                    'Dynamic option for gmail password. '
                    'If provided and email notifier is requested, will '
                    'use this to send email via gmail. '
                    'If not set, will lookup from environment.')),
    OxMonOption('--OX_MON_SES_PROFILE', envvar='OX_MON_SES_PROFILE',
                default=None, help=(
                    'Dynamic option for AWS SES profile for email. '
                    'If provided and email notifier is requested, will '
                    'use this to send email via AWS SES. '
                    'If not set, will lookup from environment.')),
    OxMonOption('--loglevel', default='INFO', type=click.Choice([
        'DEBUG', 'INFO', 'WARNING', 'CRITICAL', 'ERROR']), help=(
            'Root log level to use while running.')),
    ]


class BasicConfig:
    """Configuration object for ox_mon.
    """

    def __init__(self, **kwargs):
        for name, value in kwargs.items():
            setattr(self, name, value)



"""Common interface for various types of monitors in ox_mon
"""


import datetime
import logging

from ox_mon.common import exceptions
from ox_mon.common import noters


class OxMonTask:
    """Abstract class representing interface for ox mon task

Each task should implement the following interface:

  1. Have an `__init__` method that takes a config object.
  2. Have a _do_task method (see docs below).
  3. Have an options method (see docs below).

With those features implemented, it is easy to turn run tasks in
a consistent way by creating an instance and running the run method.
This can be done either using pure python or through the command line
as documented elsewhere.
    """

    def _do_task(self) -> str:
        """Main method to run task; raise OxMonAlarm if problem found.

This method should do whatever it needs to execute the monitoring task.
If everything is OK, it can return a string status message if desired or
just return None. If problems are found, this should raise an instance
of OxMonAlarm describing the problem.

That will be caught by the run method and a notification will be sent
to the appropriate place.

Users should call `run` which calls this method; do not call this
method directlry.
        """
        raise NotImplementedError

    @classmethod
    def options(cls):
        """Return list of configs.OxMonOption for configure this task.
        """
        raise NotImplementedError

    def __init__(self, config):
        self.config = config

    def notify(self, subject: str, msg: str):
        """Notify for given subject and message.

        :param subject:   String subject for notification.

        :param msg:       String message for notification.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:  Go through all notifiers in self.config.notifiers and
                  send the given notification message.
        """
        for ntype in self.config.notifier:
            my_noter = noters.make_notifier(ntype, self.config)
            my_noter.send(subject, msg)

    def run(self):
        """Run the _do_task method with notification, etc.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        """
        try:
            result = self._do_task()
            return result
        except exceptions.OxMonAlarm as ox_alarm:
            logging.info(
                'Raising ox_mon alarm: %s; will attempt to notify', ox_alarm)
            summary = 'ox_mon alarm for %s' % self.__class__.__name__
            self.notify(summary, '%s at UTC %s\n%s' % (
                summary, datetime.datetime.utcnow(), str(ox_alarm)))
            raise
        except exceptions.OxMonException as ox_prob:
            logging.warning(
                'Got ox_mon exception: %s; will attempt to notify', ox_prob)
            self.notify('Error: ox_mon failed',
                        'Error: ox_mon failed:\n%s' % str(ox_prob))
            raise
        except Exception as unexpected:
            logging.error(
                'Got unexpected exception %s; will attempt to notify', str(
                    unexpected))
            self.notify('Error: ox_mon unexpected exception',
                        'Error: ox_mon unexpected exception:\n%s' % str(
                            unexpected))
            raise


class Checker(OxMonTask):
    """Abstract class representing interface for Checker.

A Checker is a sub-class of OxMonTask to check something but
not really modify the system itself. Sub-classes should
implement the _check method.
    """

    def _check(self) -> str:
        """Main method to check system; raise OxMonAlarm if problem found.

This method should do whatever it needs to check the status of the system.
If everything is OK, it can return a string status message if desired or
just return None. If problems are found, this should raise an instance
of OxMonAlarm describing the problem.
        """
        raise NotImplementedError

    def _do_task(self) -> str:
        "Just return _check()"
        return self._check()



"""Module for checking ClamAV using shell commands.

"""

import re
import logging
import os
import subprocess
import tempfile

import click

from ox_mon.common import configs, interface, exceptions


class InfectedFilesFound(exceptions.OxMonAlarm):
    "Exception to indicated infected files found."

    def __init__(self, scanned, infected, summary, *args, **kwargs):
        msg = 'Found %i / %i infected files in clamscan:\n%ssummary' % (
            infected, scanned, summary)
        super().__init__(msg, *args, **kwargs)


class ClamScanShellChecker(interface.Checker):
    """Version of clamscan checker using shell commands.
    """

    @classmethod
    def options(cls):
        "Override to provide options for ClamScanShellChecker."
        logging.debug('Making options for %s', cls)
        result = configs.BASIC_OPTIONS + [
            configs.OxMonOption(
                'target', type=click.Path(exists=True), help=(
                    'Path to target file or directory to scan.')),
            configs.OxMonOption(
                'clamopts', default=':r', help=(
                    'Comma separated list of options for clamscan. '
                    'We replace colons with dashes so e.g., :r becomes -r.'))
            ]
        return result

    @staticmethod
    def _parse_log(fname):
        text = open(fname).read()
        my_re = re.compile(
            r'-+ +SCAN SUMMARY +-+.*' +
            r'(?P<scanned>Scanned files:[^\n]+)\n' +
            r'(?P<infected>Infected files:[^\n]+)\n' +
            r'.*Time: .*\n*$', re.MULTILINE | re.DOTALL)
        match = my_re.search(text)
        if not match:
            raise ValueError('Could not parse clamscan results: ...%s\n' % (
                text[-300:]))
        summary = text[match.start():match.end()]
        scanned = int(match.group('scanned').split(' ')[-1])
        infected = int(match.group('infected').split(' ')[-1])
        return scanned, infected, summary

    def _check(self):
        "Check for viruses as required by parent class."

        try:
            my_fd, log = tempfile.mkstemp(suffix='_clamscan.txt')
            os.close(my_fd)
            cmd = ['clamscan', '--log=%s' % log, self.config.target]
            if self.config.clamopts:
                for item in self.config.clamopts.split(','):
                    cmd.append(item.replace(':', '-'))
            proc = subprocess.run(cmd, check=False)
            if not os.path.exists(log):
                raise ValueError('Could not find clamscan log at %s' % log)
            scanned, infected, summary = self._parse_log(log)
            if infected:
                raise InfectedFilesFound(scanned, infected, summary)
            if proc.returncode != 0:
                raise ValueError('Bad return code %s from clamscan:\n%s' % (
                    proc.returncode, summary))
            return 'No infected files from clamscan:\n%s' % summary
        finally:
            if os.path.exists(log):
                os.remove(log)



"""Tasks which are really commands.
"""

import sys
import logging
import subprocess

from ox_mon.common import configs, interface


class RawCmd(interface.OxMonTask):
    """Raw command with benefits of ox_mon notifiers, etc.

This command is helpful in case you want to run a shell script
or other basic command and use ox_mon to verify that it ran correclty
with notifications sent if there are problems.
    """

    @classmethod
    def options(cls):
        logging.debug('Making options for %s', cls)
        result = configs.BASIC_OPTIONS + [
            configs.OxMonOption(
                '--cmd', help=('Command to run.')),
            configs.OxMonOption(
                '--args', default='', help=(
                    'Comma separated list of args to pass to --cmd. '
                    'We replace colons with dashes so e.g., :v becomes -v.')),
            configs.OxMonOption(
                '--stdout', default='@STDOUT', help=(
                    'Where to send stdout of the underling command. '
                    'If you provide @STDOUT then the output from --cmd '
                    'will go on the standard output stream. You can also '
                    'provide @STDERR or path to a file or @NULL.')),
            configs.OxMonOption(
                '--stderr', default='@STDERR', help=(
                    'Where to send stderr of the underling command. '
                    'If you provide @STDERR then the output from --cmd '
                    'will go on the standard error stream. You can also '
                    'provide @STDERR or path to a file or @NULL.')),
            configs.OxMonOption(
                '--shell', default=False, help=(
                    'Whether to run --cmd through the shell.')),
            ]
        return result

    @staticmethod
    def _make_out_stream(place: str):
        """Make an output stream from string description.

        :param place:   String which is either path to a file, @STDOUT,
                        @STDERR, or @NULL.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        :return:   What should be passed to subprocess.run to make the
                   output go to desired stream.

        ~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-~-

        PURPOSE:   Prepare output stream for subprocess.run

        """
        if place == '@STDOUT':
            return sys.stdout
        if place == '@STDERR':
            return sys.stderr
        if place == '@NULL':
            return subprocess.PIPE
        result = open(place, 'w')
        return result

    def _do_sub_cmd(self):
        if not self.config.cmd:
            raise ValueError('No value provided for --cmd.')
        cmd = [self.config.cmd]
        for item in self.config.args.split(','):
            cmd.append(item.replace(':', '-'))
        stdout = self._make_out_stream(self.config.stdout)
        stderr = self._make_out_stream(self.config.stderr)
        try:
            proc = subprocess.run(cmd, check=False, stdout=stdout,
                                  stderr=stderr, shell=self.config.shell)
            logging.debug('Return value for cmd "%s": %s',
                          cmd, proc.returncode)
            if proc.returncode != 0:
                stdout = proc.stdout
                if hasattr(stdout, 'decode'):
                    stdout = stdout.decode('utf8')
                stderr = proc.stderr
                if hasattr(stderr, 'decode'):
                    stderr = stderr.decode('utf8')

                raise ValueError('Bad return code %s from %s:\n%s' % (
                    proc.returncode, cmd, (
                        stdout if stdout else '') + '\n' + (
                            stderr if stderr else '')))
        finally:
            for item in [stdout, stderr]:
                if item not in [None, sys.stdout, sys.stderr] and hasattr(
                        item, 'close'):
                    item.close()

    def _do_task(self):
        return self._do_sub_cmd()



"""Stubs for various purposes.
"""

import doctest
import logging


class GenericModuleStub:
    """Generic stub for module we could not load.

Imagine that you need some module `fake_mod` for your program but only
for certain features. You want to import it and use it if that feature
is used, but not complain about import errors if the module is missing
and the feature is not used.

This class lets you create a generic module stub for that case as
illustrated below.

Basically, you catch the import error for importing your fake module,
then create a GenericModuleStub instance for the module:

>>> from ox_mon.common import stubs
>>> try:
...     import fake_mod
... except Exception as prob:
...     fake_mod = stubs.GenericModuleStub(
...         'fake_mod', str(prob), 'magical dependancies')
...
>>> try:  # show what would happen if you tried using fake_mod.bar
...     fake_mod.bar()
... except Exception as prob:
...     print(prob)
Unable to call method bar in module fake_mod.
Module fake_mod unimported because No module named 'fake_mod'.
If you want to use this feature; please install required
dependancies such as "magical dependancies"


You can also do fancier things to control exactly what happens for
certain stub method:


>>> from ox_mon.common import stubs
>>> try:
...     import fake_mod
... except Exception as prob:
...     fake_mod = stubs.GenericModuleStub(
...         'fake_mod', str(prob), 'magical dependancies')
...     fake_mod.test = fake_mod.make_complainer('test')  # stub method 'test'
...     fake_mod.foo = fake_mod.make_complainer('foo', note=(
...        'you can provide notes like: you really do not need foo anyway'))
...
>>> try:
...     fake_mod.test()
... except Exception as prob:
...     print(prob)
Unable to call method test in module fake_mod.
Module fake_mod unimported because No module named 'fake_mod'.
If you want to use this feature; please install required
dependancies such as "magical dependancies"


>>> try:
...     fake_mod.foo()
... except Exception as prob:
...     print(prob)
Unable to call method foo in module fake_mod.
Module fake_mod unimported because No module named 'fake_mod'.
If you want to use this feature; please install required
dependancies such as "magical dependancies"
you can provide notes like: you really do not need foo anyway

    """

    def __init__(self, modname, ierr, deps):
        """Initializer.

        :param modname:    String name of attempted module load.

        :param ierr:       String error in importing.

        :param deps:       String list of dependancies.

        """
        self.modname = modname
        self.ierr = ierr
        self.deps = deps

    def __getattr__(self, name):
        complainer = self.make_complainer(name)
        return complainer

    def complaint_msg(self, method: str, note: str = None) -> str:
        """Take in string method name and return complaint for it.
        """
        msg = '\n'.join([
            'Unable to call method %s in module %s.' % (method, self.modname),
            'Module %s unimported because %s.' % (self.modname, self.ierr),
            'If you want to use this feature; please install required',
            'dependancies such as "%s"' % self.deps] + (
                [note] if note else []))

        return msg

    def make_complainer(self, method_name: str, note: str = ''):
        """Take method name and return callable for it.

See class docs for example usage.
        """
        msg = self.complaint_msg(method_name, note)

        def complainer(*args, **kwargs):
            "Return function to complain if called."
            logging.debug('Ignoring *args=%s, **kwargs=%s for %s',
                          args, kwargs, method_name)
            raise ValueError(msg)
        return complainer


if __name__ == '__main__':
    doctest.testmod()
    print('Finished Tests')
